﻿/**
 * Angle Mouse
 *
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *		
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2009-2011, Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.Diagnostics;
using System.Resources;
using System.Reflection;
using WobbrockLib;
using WobbrockLib.Devices;
using WobbrockLib.Extensions;
using WobbrockLib.Controls;

namespace AngleMouse
{
    /// <summary>
    /// 
    /// </summary>
    [Serializable]
    public partial class MainForm : Form, ISerializable
    {
        #region Members

        private const long TDelayReset = 500; // ms
        public const string SerializationFile = "\\settings.ini";
        private static readonly double[] InitialWeights = new double[2] { 0.5, 0.5 }; // even weighting for two points

        private uint _savedGain;
        private bool _savedUsePA;
        private bool _showedTrayTipThisRun;

        private LowLevelMouseHook _mhook;
        private List<PointF> _moves;
        private long _tLastMove;
        private double[] _weights;
        private double _maxAngularStdev;

        private Options _o;

        #endregion

        #region Start & Stop

        /// <summary>
        /// Default constructor.
        /// </summary>
        public MainForm()
        {
            _o = new Options(PresetType.Some);
            CtorCommon(); // put after _o is set
        }

        /// <summary>
        /// Deserialization constructor.
        /// </summary>
        /// <param name="info">Serialization information.</param>
        /// <param name="context">Streaming context. Currently unused.</param>
        protected MainForm(SerializationInfo info, StreamingContext context)
        {
            _o = new Options(info, context);
            CtorCommon(); // put after _o is set

            this.Left = info.GetInt32("Left");
            this.Top = info.GetInt32("Top");
            this.Width = info.GetInt32("Width");
            this.Height = info.GetInt32("Height");

            chkHide.Checked = info.GetBoolean("HideWindow");
        }

        /// <summary>
        /// Common constructor shared between default and serialization constructors.
        /// </summary>
        private void CtorCommon()
        {
            InitializeComponent();

            _showedTrayTipThisRun = false;
            _moves = new List<PointF>(_o.QueueLength);
            _weights = InitialWeights;
            _tLastMove = 0;

            _mhook = new LowLevelMouseHook("Angle Mouse Hook");
            _mhook.OnMouseMove += new EventHandler<MouseEventArgs>(OnHookMouseMove);
            _mhook.OnMouseClick += new EventHandler<MouseEventArgs>(OnHookMouseClick);
        }

        /// <summary>
        /// Writes the object to the given file.
        /// </summary>
        /// <param name="file">The file to which to write this object.</param>
        public void Serialize(string file)
        {
            Stream stream = null;
            try
            {
                IFormatter formatter = new BinaryFormatter();
                stream = new FileStream(file, FileMode.Create, FileAccess.Write);
                formatter.Serialize(stream, this);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
            finally
            {
                if (stream != null)
                    stream.Close();
            }
        }

        /// <summary>
        /// Read in the object from the given file.
        /// </summary>
        /// <param name="file">The file from which to deserialize this object.</param>
        /// <returns>The inflated MainForm form, read from the given file, or null if the
        /// file does not exist.</returns>
        public static MainForm Deserialize(string file)
        {
            MainForm frm = null;
            Stream stream = null;
            try
            {
                IFormatter formatter = new BinaryFormatter();
                stream = new FileStream(file, FileMode.Open, FileAccess.Read);
                frm = (MainForm) formatter.Deserialize(stream);
            }
            catch (FileNotFoundException fnfex)
            {
                Console.WriteLine(fnfex);
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex);
            }
            finally
            {
                if (stream != null)
                    stream.Close();
            }
            return frm;
        }

        /// <summary>
        /// Called during serialization. Saves the data out for reading in later.
        /// </summary>
        /// <param name="info">Serialization information. Holds whatever properties we save into it.</param>
        /// <param name="context">Serialization context. Currently unused.</param>
        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            // have all Options write themselves out
            _o.GetObjectData(info, context);

            // then write out our form location settings
            info.AddValue("Left", this.Left);
            info.AddValue("Top", this.Top);
            info.AddValue("Width", this.Width);
            info.AddValue("Height", this.Height);

            // and finally write out our hide window setting
            info.AddValue("HideWindow", chkHide.Checked);
        }

        /// <summary>
        /// Load event handler for the main form.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void MainForm_Load(object sender, EventArgs e)
        {
            // save and set the gain
            _savedGain = ControlPanel.MouseGain; // save
            _savedUsePA = ControlPanel.EnhancePointingPrecision; // save

            // set a preset if one is matched by our settings
            if (CheckForPreset() == PresetType.None)
            {
                // turn on or off mouse hook (there was no preset that did it for us)
                SetMouseHook(_o.IsActive);
            }

            // hide the window on startup?
            if (chkHide.Checked)
            {
                cmdTray_Click(sender, e); // as if the form was sent immediately to the system tray
                _showedTrayTipThisRun = false; // unset -- still show tip if system tray button is clicked later
            }
            else
            {
                this.BringToFront();
            }
        }

        /// <summary>
        /// Write out the preferences for the application.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (this.WindowState == FormWindowState.Minimized)
            {
                this.WindowState = FormWindowState.Normal; // restore before serializing
            }
            Serialize(Application.StartupPath + SerializationFile); // write out
        }

        /// <summary>
        /// FormClosed event handler for the main form.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void MainForm_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (!SetMouseHook(false)) // turn off mouse hook
            {
                ControlPanel.MouseGain = _savedGain; // ensure restored
                ControlPanel.EnhancePointingPrecision = _savedUsePA; // ensure restored
            }
        }

        #endregion

        #region Angle Mouse Logic

        /// <summary>
        /// When the mouse is clicked, clear the moves (and therefore, angle queue), and restore
        /// the intitial weights for the Angle Mouse. We also clear the text box outputs and
        /// the graph showing the current angle queue weights.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void OnHookMouseClick(object sender, MouseEventArgs e)
        {
            _moves.Clear();
            _weights = InitialWeights;
            _tLastMove = 0;

            txtAngles.Clear();
            txtMean.Clear();
            txtDeviation.Clear();
            txtMaxDeviation.Clear();
            txtPercent.Clear();
            txtMouseSpeed.Clear();

            if (e.Button != MouseButtons.Right)
            {
                grfCurve.ClearSeries();
                grfCurve.Invalidate();
            }
        }

        /// <summary>
        /// This is the main entrypoint for mouse movement in the Angle Mouse. This method is the
        /// essence of the research described in the CHI 2009 Angle Mouse paper by Wobbrock et al.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void OnHookMouseMove(object sender, MouseEventArgs e)
        {
            // test whether enough time has elapsed since the last move
            long now = TimeEx.NowMs;
            if (_tLastMove != 0 && now - _tLastMove > TDelayReset)
                OnHookMouseClick(sender, e); // reset as if a click happened
            _tLastMove = now;

            // test whether the move point is threshold distance from the last point
            PointF p = new PointF(e.X, e.Y);
            if (_moves.Count == 0)
            {
                _moves.Add(p);
            }
            else
            {
                PointF p0 = _moves[0]; // last move
                double dx = GeotrigEx.Distance(p0, p);
                if (dx >= _o.DistanceThreshold) // was it far enough?
                {
                    if (_moves.Count == _o.QueueLength) // is our queue full?
                        _moves.RemoveAt(_moves.Count - 1); // remove tail
                    _moves.Insert(0, p); // insert new head

                    if (_moves.Count >= 3) // need 3+ moves for 2+ angles to have any angular deviation ("spread")
                    {
                        // get the array of current angles
                        double[] angles = GetAngles(_moves);

                        // compute the weighted angular mean (Eq. 9)
                        double wmean = GetAngularMean(angles, _weights);

                        // compute the weighted angular standard deviation (Eqs. 7-8)
                        double wstdev = GetAngularStdev(angles, _weights, wmean);

                        // map the weighted angular standard deviation to the Gaussian standard deviation (Eq. 12)
                        double maxstdev = GetMaxAngularStdev(angles.Length, _weights);
                        double pctstdev = GeotrigEx.PinValue(wstdev / maxstdev, 0.0, 1.0); // percent of max possible stdev
                        double gausstdev = GeotrigEx.MapValue(pctstdev, 0.00, 1.00, _o.MinGaussianStdev, _o.MaxGaussianStdev, true);

                        // recalculate our weights based on the new Gaussian standard deviation (Eq. 11)
                        _weights = GetAngleWeights(angles.Length, gausstdev);

                        // compute our weighted angular standard deviation (Eqs. 7-9)
                        wmean = GetAngularMean(angles, _weights);
                        wstdev = GetAngularStdev(angles, _weights, wmean);

                        // map our weighted angular standard deviation to our gain (Eq. 10)
                        maxstdev = GetMaxAngularStdev(angles.Length, _weights);
                        pctstdev = GeotrigEx.PinValue(1.0 - wstdev / maxstdev, 0.0, 1.0); // percent of max possible stdev 
                        double gain = GeotrigEx.MapValue(pctstdev, 0.00, 1.00, _o.MinGain, _o.MaxGain, true);

                        // set the gain
                        ControlPanel.MouseGain = (uint) Math.Round(gain);
                        
                        // recompute our weights in preparation for the next movement
                        _weights = GetAngleWeights(Math.Min(angles.Length + 1, _o.QueueLength - 1), gausstdev);

                        // update all the text boxes for output
                        txtAngles.AppendText(String.Format("\r\n{0}", _moves.Count - 1));
                        txtMean.AppendText(String.Format("\r\n{0:f2}°", GeotrigEx.Radians2Degrees(wmean)));
                        txtDeviation.AppendText(String.Format("\r\n{0:f2}°", GeotrigEx.Radians2Degrees(wstdev)));
                        txtMaxDeviation.AppendText(String.Format("\r\n{0:f2}°", GeotrigEx.Radians2Degrees(maxstdev)));
                        txtPercent.AppendText(String.Format("\r\n{0:f2}", pctstdev));
                        txtMouseSpeed.AppendText(String.Format("\r\n{0:f2}", gain));

                        txtAngles.ScrollToCaret();
                        txtMean.ScrollToCaret();
                        txtDeviation.ScrollToCaret();
                        txtMaxDeviation.ScrollToCaret();
                        txtPercent.ScrollToCaret();
                        txtMouseSpeed.ScrollToCaret();

                        // update the graph of the current curve being used
                        GraphCurrentCurve(angles.Length, gausstdev);
                    }
                }
            }
        }
        
        /// <summary>
        /// Graph the current angular weighting curve used in the Angle Mouse. The user can see this 
        /// curve peaking when movement is relatively coherent, and flattening when movement is relatively
        /// divergent.
        /// </summary>
        /// <param name="numAngles">The number of angles currently in the angle queue.</param>
        /// <param name="gausstdev">The current dynamic Gaussian standard deviation used to configure the curve.</param>
        private void GraphCurrentCurve(int numAngles, double gausstdev)
        {
            grfCurve.XAxisTicks = numAngles - 1;
            grfCurve.ClearSeries();

            Graph.Series xaxis = new Graph.Series("X-Axis", Color.Black, Color.Black, false, false);
            xaxis.AddPoint(1f, 0f);
            xaxis.AddPoint(numAngles, 0f);
            grfCurve.AddSeries(xaxis);

            Graph.Series yaxis = new Graph.Series("Y-Axis", Color.Black, Color.Black, false, false);
            yaxis.AddPoint(1f, 0f);
            yaxis.AddPoint(1f, 0.30f);
            grfCurve.AddSeries(yaxis);

            PointF[] gpts = new PointF[numAngles];
            float gsum = 0f;
            for (int j = 1; j <= numAngles; j++)
            {
                double g = (1.0 / (gausstdev * Math.Sqrt(2.0 * Math.PI))) * Math.Pow(Math.E, -(j * j) / (2.0 * gausstdev * gausstdev));
                gpts[j - 1] = new PointF(j, (float) g);
                gsum += (float) g;
            }
            for (int j = 0; j < numAngles; j++)
            {
                gpts[j].Y /= gsum; // normalize
            }
            Graph.Series s = new Graph.Series(String.Format("stdev={0:f2}", gausstdev), Color.Black, RandomEx.Color(0, 200), false, true);
            s.AddPoints(gpts);
            grfCurve.AddSeries(s);
            grfCurve.Invalidate();
        }

        /// <summary>
        /// Computes the angles formed by the successive movement points.
        /// </summary>
        /// <returns>The angles between successive points.</returns>
        private double[] GetAngles(List<PointF> pts)
        {
            double[] angles = new double[pts.Count - 1];
            for (int i = 1; i < pts.Count; i++)
            {
                angles[i - 1] = GeotrigEx.Angle(pts[i], pts[i - 1], true);
            }
            return angles;
        }

        /// <summary>
        /// Computes the weighted mean angle defined as the angle between all others.
        /// For example, the mean angle between 359 degrees and 1 degree is 0 degrees.
        /// Thus, this is not a simple arithmetic calculation, as (359+1)/2 = 180.
        /// Instead, it is the mean unit vector of angles given.
        /// </summary>
        /// <param name="angles">The angles from which to calculate the mean.</param>
        /// <param name="wgts">The weights to use in computing the weighted mean.</param>
        /// <returns>The weighted angular mean.</returns>
        public double GetAngularMean(double[] angles, double[] wgts)
        {
            double[] x = new double[angles.Length]; // x-coords of unit vectors
            double[] y = new double[angles.Length]; // y-coords of unit vectors

            // store up the unit-vector (x,y) coordinates
            for (int i = 0; i < angles.Length; i++)
            {
                PointF p = GeotrigEx.PointFromVector(PointF.Empty, 1.0f, angles[i]); // unit vector
                x[i] = p.X;
                y[i] = p.Y;
            }

            // take the weighted means of the unit vector coordinates
            double x_bar = StatsEx.Mean(x, wgts);
            double y_bar = StatsEx.Mean(y, wgts);

            // get angle from (0,0) to (x_bar, y_bar)
            PointF p_bar = new PointF((float) x_bar, (float) y_bar);
            return GeotrigEx.Angle(PointF.Empty, p_bar, true); // weighted angular mean
        }

        /// <summary>
        /// Computes the weighted standard deviation of angles.
        /// </summary>
        /// <param name="angles">Angles in radians whose standard deviation to compute.</param>
        /// <param name="wgts">The weights to use in the weighted calculation.</param>
        /// <param name="wmean">The weighted mean angle to use. (We could compute that within this method
        /// but it would be redundant since we compute it previously.)</param>
        /// <returns>The weighted standard deviation of angles.</returns>
        /// <remarks>WobbrockLib.Extensions.StatsEx implements a weighted standard deviation
        /// but we cannot use it because the deviation from the mean must be an angular distance
        /// calculation, not a weighted arithmetic mean.</remarks>
        public double GetAngularStdev(double[] angles, double[] wgts, double wmean)
        {
            if (angles.Length <= 1)
                return 0.0; // no deviation with one angle

            double sumw = 0.0; // sum of used weights
            double sumsqw = 0.0; // sum of squared weights
            double sumwsqdev = 0.0; // sum of weighted squared deviations

            for (int i = 0; i < angles.Length; i++)
            {
                double da = GeotrigEx.RadiansBetween(wmean, angles[i]); // Eqs. 2-3, 8
                sumwsqdev += wgts[i] * (da * da); // sum of weighted squared deviations
                sumsqw += wgts[i] * wgts[i]; // sum of squared weights
                sumw += wgts[i]; // keep sum of weights
            }

            double variance = (sumw / (sumw * sumw - sumsqw)) * sumwsqdev;
            return Math.Sqrt(variance); // wstdev, Eq. 7
        }

        /// <summary>
        /// Create weights such that more recent angles matter more than angles from long ago.
        /// </summary>
        /// <param name="numAngles">The number of angles for which to create weights.</param>
        /// <param name="gausstdev">The Gaussian standard deviation to use when creating te weights.</param>
        /// <returns>A weighting kernel whose weights sum to 1.00.</returns>
        private double[] GetAngleWeights(int numAngles, double gausstdev)
        {
            double[] k = new double[numAngles];

            // create the kernel values, only discretizing the slots needed for a half-Gaussian
            double ksum = 0.0;
            for (int i = 1; i <= k.Length; i++)
            {
                k[i - 1] = (1.0 / (gausstdev * Math.Sqrt(2.0 * Math.PI))) * Math.Pow(Math.E, -(i * i) / (2.0 * gausstdev * gausstdev)); // Eq. 11
                ksum += k[i - 1]; // track the sum
            }
            // the sum in the kernel may not always be 1.0. we can normalize each kernel value by 
            // dividing it by the sum of all weights in the kernel. then the kernel will add to 1.0.
            for (int i = 0; i < k.Length; i++)
            {
                k[i] /= ksum;
            }
            return k;
        }

        /// <summary>
        /// Estimate the maximum possible angular deviation by choosing random angles.
        /// Remember the maximum we have seen and if the randomly generated deviation 
        /// is less, return the previous maximum. If it is more, update and reutrn the new
        /// maximum.
        /// </summary>
        /// <param name="numAngles">The number of angles.</param>
        /// <param name="wgts">The weights on each angle.</param>
        /// <returns>An estimate for the maximum possible angular deviation. The more angles that
        /// are used, the better the estimate.</returns>
        /// <remarks>We cannot use WobbrockLib.Extensions.StatsEx.MaxStdDev because the deviation 
        /// from the mean must be an angular distance calculation, not a weighted arithmetic mean.</remarks>
        private double GetMaxAngularStdev(int numAngles, double[] wgts)
        {
            double[] randangles = new double[numAngles];
            for (int i = 0; i < randangles.Length; i++)
            {
                int deg = RandomEx.Integer(0, 359);
                randangles[i] = GeotrigEx.Degrees2Radians(deg);
            }
            double wmean = GetAngularMean(randangles, wgts);
            double wstdev = GetAngularStdev(randangles, wgts, wmean);
            if (wstdev > _maxAngularStdev) // if max is unset at 0.0, this will always pass
                _maxAngularStdev = wstdev; // update
            return _maxAngularStdev;
        }

        #endregion

        #region Tab Control

        /// <summary>
        /// When the Presets tab is selected, check to see if the Custom settings equal
        /// any of the defined Presets. If so, highlight the Preset accordingly. No need to
        /// do this whenever any settings change, just when the tab is revealed.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void tabMain_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (tabMain.SelectedIndex == 0) // presets
            {
                CheckForPreset();
            }
            else if (tabMain.SelectedIndex == 1) // custom
            {
                UpdateCustomWidgets();
            }
        }

        #endregion

        #region Presets Tab

        /// <summary>
        /// Event handler for all of the option (radio) buttons on the Presets tab. The entered
        /// button has its border turned yellow as roll-over feedback.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void optPreset_MouseEnter(object sender, EventArgs e)
        {
            RadioButton r = (RadioButton) sender;
            r.FlatAppearance.BorderColor = Color.Yellow;
        }

        /// <summary>
        /// Event handler for all of the option (radio) buttons on the Presets tab. The departed
        /// button has its border turned white to end roll-over feedback, or orange if that preset
        /// is currently selected.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void optPreset_MouseLeave(object sender, EventArgs e)
        {
            RadioButton r = (RadioButton) sender;
            r.FlatAppearance.BorderColor = (r.Checked ? Color.Orange : Color.White);
        }

        /// <summary>
        /// Event handler for all of the option (radio) buttons on the Presets tab. The option
        /// button whose checked state has changed will change the current widget settings on
        /// the Custom tab to reflect the current state of the settings.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void optPreset_CheckedChanged(object sender, EventArgs e)
        {
            RadioButton r = (RadioButton) sender;
            int itag = Int32.Parse((string) r.Tag);
            if (itag != 0) // deactivate
                r.FlatAppearance.BorderColor = (r.Checked ? Color.Orange : Color.White);

            if (r.Checked) // if something was checked (and not unchecked)
            {
                // make Angle Mouse presets take effect here
                if (itag == 0) // deactivate
                {
                    _o.IsActive = false;
                }
                else
                {
                    _o = new Options((PresetType) itag); // options for this preset
                    _o.IsActive = true; // all presets assume the Angle Mouse is on
                }
                SetMouseHook(_o.IsActive);
            }
        }

        /// <summary>
        /// Checks to see whether the current settings are equal to any preset settings.
        /// If so, the proper preset radio button is checked. If the Angle Mouse is not
        /// currently active, then regardless of the current settings, the Deactivate
        /// radio button will be checked.
        /// </summary>
        /// <remarks>If these setters change the checked state of the widget, the optPreset_CheckedChanged
        /// event for that widget will fire.</remarks>
        private PresetType CheckForPreset()
        {
            if (_o.IsActive)
            {
                PresetType p = _o.TestForPreset();
                switch (p)
                {
                    case PresetType.None: // uncheck all
                        optHardlyAny.Checked = false;
                        optAFew.Checked = false;
                        optSome.Checked = false;
                        optMany.Checked = false;
                        optWayTooMany.Checked = false;
                        optDeactivate.Checked = false;
                        break;
                    case PresetType.HardlyAny:
                        optHardlyAny.Checked = true;
                        break;
                    case PresetType.AFew:
                        optAFew.Checked = true; 
                        break;
                    case PresetType.Some:
                        optSome.Checked = true; 
                        break;
                    case PresetType.Many:
                        optMany.Checked = true;
                        break;
                    case PresetType.WayTooMany:
                        optWayTooMany.Checked = true;
                        break;
                }
                return p;
            }
            else optDeactivate.Checked = true;
            return PresetType.None;
        }

        #endregion

        #region Custom Tab

        /// <summary>
        /// When the checkbox on the Custom tab for activating the Angle Mouse is
        /// changed, this event handler fires. Our options state is updated and
        /// the mouse hook is set.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void chkActivate_CheckedChanged(object sender, EventArgs e)
        {
            _o.IsActive = chkActivate.Checked;
            SetMouseHook(_o.IsActive);
        }

        /// <summary>
        /// When the checkbox on the Custom tab for activating Pointer Acceleration is
        /// changed, this event handler fires. Our options state is updated and
        /// the mouse control panel setting is enabled.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void chkUsePA_CheckedChanged(object sender, EventArgs e)
        {
            _o.UsePointerAcceleration = chkUsePA.Checked;
            ControlPanel.EnhancePointingPrecision = chkUsePA.Checked;
        }

        /// <summary>
        /// Handles when the distance threshold for movement that produces another angle is set.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numDistanceThreshold_ValueChanged(object sender, EventArgs e)
        {
            _o.DistanceThreshold = (double) numDistanceThreshold.Value;
        }

        /// <summary>
        /// Handles the change in the movement queue length parameter. To help the user visualize the effects
        /// of this change, we graph the Gaussian weighting curves, with one curve for each integer Gaussian
        /// standard deviation. The user's goal should be to have the most peaked curve reach zero just about
        /// the final angle queue number, and the least peaked curve to be reasonably flat. The most peaked 
        /// curve will be used when mouse movement is reasonably coherent, while the least peaked curve will
        /// create semi-constant weighting, and be used when the mouse movement is reasonably divergent.
        /// Using the most peaked curve when movement is coherent means that only the most recent angles 
        /// contribute much to the angular deviation calculation. Using the least peaked curve when 
        /// movement is divergent means that all angles are "remembered," contributing about evenly, in the
        /// case of a fairly flat curve, to the angular deviation calculation.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numQueueLength_ValueChanged(object sender, EventArgs e)
        {
            _o.QueueLength = (int) numQueueLength.Value;
            _moves.Clear(); // reset
            _maxAngularStdev = 0.0; // reset
            UpdateGraphOfWeights(_o.QueueLength - 1, _o.MinGaussianStdev, _o.MaxGaussianStdev);
        }

        /// <summary>
        /// The maximum Gaussian standard deviation widget determines how flat the weighting curve
        /// is for when movement is highly divergent.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numMaxGauss_ValueChanged(object sender, EventArgs e)
        {
            _o.MaxGaussianStdev = (int) numMaxGauss.Value;
            numMinGauss.Maximum = numMaxGauss.Value;
            _moves.Clear(); // reset
            _maxAngularStdev = 0.0; // reset
            UpdateGraphOfWeights(_o.QueueLength - 1, _o.MinGaussianStdev, _o.MaxGaussianStdev);
        }

        /// <summary>
        /// The minimum Gaussian standard deviation widget determines how peaked the weighting curve
        /// is for when movement is highly coherent.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numMinGauss_ValueChanged(object sender, EventArgs e)
        {
            _o.MinGaussianStdev = (int) numMinGauss.Value;
            numMaxGauss.Minimum = numMinGauss.Value;
            _moves.Clear(); // reset
            _maxAngularStdev = 0.0; // reset
            UpdateGraphOfWeights(_o.QueueLength - 1, _o.MinGaussianStdev, _o.MaxGaussianStdev);
        }

        /// <summary>
        /// The widget handler for the max C-D gain slider setting from 1-20, as defined by the Windows API.
        /// This value reflects the fastest possible mouse speed.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numMaxGain_ValueChanged(object sender, EventArgs e)
        {
            _o.MaxGain = (uint) numMaxGain.Value;
            numMinGain.Maximum = numMaxGain.Value;
        }

        /// <summary>
        /// The widget handler for the min C-D gain slider setting from 1-20, as defined by the Windows API.
        /// This value reflects the slowest possible mouse speed.
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void numMinGain_ValueChanged(object sender, EventArgs e)
        {
            _o.MinGain = (uint) numMinGain.Value;
            numMaxGain.Minimum = numMinGain.Value;
        }

        /// <summary>
        /// Turns on or off the mouse hook, which observes mouse movement regardless of which application
        /// the user is in.
        /// </summary>
        /// <param name="on">If true, turns on the hook; if false, turns it off.</param>
        /// <returns>True if the call actually did something (took effect); false otherwise.</returns>
        private bool SetMouseHook(bool on)
        {
            bool didSomething = false;
            if (on && !_mhook.IsActive) // turn on the hook
            {
                ControlPanel.MouseGain = _o.MaxGain; // set
                ControlPanel.EnhancePointingPrecision = _o.UsePointerAcceleration; // set
                _mhook.Install(); // install
                didSomething = true;
            }
            else if (!on && _mhook.IsActive) // or turn it off
            {
                _mhook.Uninstall(); // uninstall
                ControlPanel.MouseGain = _savedGain; // restore
                ControlPanel.EnhancePointingPrecision = _savedUsePA; // restore
                didSomething = true;
            }
            return didSomething;
        }

        /// <summary>
        /// Load the values from the current Options settings into the widgets on the Custom tab.
        /// Many of these may cause change events to fire for the given widgets.
        /// </summary>
        private void UpdateCustomWidgets()
        {
            // keep the min/max numeric boxes from being able to cross over each other
            numMaxGauss.Minimum = _o.MinGaussianStdev;
            numMinGauss.Maximum = _o.MaxGaussianStdev;
            numMaxGain.Minimum = _o.MinGain;
            numMinGain.Maximum = _o.MaxGain;

            // load up UI controls with current state
            chkActivate.Checked = _o.IsActive;
            numDistanceThreshold.Value = (decimal) _o.DistanceThreshold;
            numQueueLength.Value = _o.QueueLength;
            numMaxGauss.Value = _o.MaxGaussianStdev;
            numMinGauss.Value = _o.MinGaussianStdev;
            numMaxGain.Value = _o.MaxGain;
            numMinGain.Value = _o.MinGain;
            chkUsePA.Checked = _o.UsePointerAcceleration;

            // update the drawn graph of weights
            UpdateGraphOfWeights(_o.QueueLength - 1, _o.MinGaussianStdev, _o.MaxGaussianStdev);
        }

        /// <summary>
        /// Helper function to create the series and update the graph of dynamic
        /// Gaussian angle weights.
        /// </summary>
        /// <param name="numAngles">The number of angles in the angle queue.</param>
        /// <param name="minGauss">The minimum Gaussian standard deviation used.</param>
        /// <param name="maxGauss">The maximum Gaussian standard deviation used.</param>
        private void UpdateGraphOfWeights(int numAngles, int minGauss, int maxGauss)
        {
            grfWeights.XAxisTicks = numAngles - 1;
            grfWeights.ClearSeries();

            Graph.Series xaxis = new Graph.Series("X-Axis", Color.Black, Color.Black, false, false);
            xaxis.AddPoint(1f, 0f);
            xaxis.AddPoint(numAngles, 0f);
            grfWeights.AddSeries(xaxis);

            Graph.Series yaxis = new Graph.Series("Y-Axis", Color.Black, Color.Black, false, false);
            yaxis.AddPoint(1f, 0f);
            yaxis.AddPoint(1f, 0.30f);
            grfWeights.AddSeries(yaxis);

            for (int i = minGauss; i <= maxGauss; i++)
            {
                PointF[] gpts = new PointF[numAngles];
                float gsum = 0f;
                for (int j = 1; j <= numAngles; j++)
                {
                    double g = (1.0 / (i * Math.Sqrt(2.0 * Math.PI))) * Math.Pow(Math.E, -(j * j) / (2.0 * i * i));
                    gpts[j - 1] = new PointF(j, (float) g);
                    gsum += (float) g;
                }
                for (int j = 0; j < numAngles; j++)
                {
                    gpts[j].Y /= gsum; // normalize
                }
                Graph.Series s = new Graph.Series(String.Format("stdev={0}", i), Color.Black, RandomEx.Color(0, 200), false, true);
                s.AddPoints(gpts);
                grfWeights.AddSeries(s);
            }
            grfWeights.Invalidate();
        }

        #endregion

        #region Button Event Handlers

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void cmdMouse_Click(object sender, EventArgs e)
        {
            ControlPanel.OpenMouseControlPanel(2);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void cmdTray_Click(object sender, EventArgs e)
        {
            this.WindowState = FormWindowState.Minimized;
            this.ShowInTaskbar = false;
            if (!_showedTrayTipThisRun)
            {
                sysIcon.ShowBalloonTip(2000, "The Angle Mouse is here.", "The Angle Mouse is still running and can be accessed via this System Tray icon.", ToolTipIcon.Info);
                _showedTrayTipThisRun = true; // only show the tip once per run
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void cmdAbout_Click(object sender, EventArgs e)
        {
            AboutForm frm = new AboutForm();
            frm.ShowDialog(this);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void cmdExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        #endregion

        #region System Tray Icon and System Tray Menu

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void sysIcon_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (tabMain.SelectedIndex == 0)
                {
                    CheckForPreset();
                }
                this.ShowInTaskbar = true;
                this.WindowState = FormWindowState.Normal;
                this.BringToFront();
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void cmnuMain_Opening(object sender, CancelEventArgs e)
        {
            mniActive.Checked = _o.IsActive;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void mniActive_Click(object sender, EventArgs e)
        {
            _o.IsActive = !_o.IsActive;
            chkActivate.Checked = _o.IsActive; // triggers chkActivate_CheckedChanged
            if (tabMain.SelectedIndex == 0)
            {
                CheckForPreset();
            }
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mniOpen_Click(object sender, EventArgs e)
        {
            if (tabMain.SelectedIndex == 0)
            {
                CheckForPreset();
            }
            this.ShowInTaskbar = true;
            this.WindowState = FormWindowState.Normal;
            this.BringToFront();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void mniExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        #endregion

        #region Explanatory Hyperlink Labels: Presets Tab

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lklblHardlyAny_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Choose this preset if you hardly make any corrections when pointing. You probably do not\r\nneed the Angle Mouse, but it will not harm your pointing performance and may improve it\r\nslightly in some cases.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lklblAFew_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Choose this preset if you often make a few corrections when pointing. The Angle Mouse\r\nwill have only a subtle effect, but should make small targets easier to acquire.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lklblSome_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Choose this preset if you often make some corrections when pointing. The Angle Mouse\r\nwill not dramatically affect the mouse, but it will make small targets easier to acquire.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lklblMany_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Choose this preset if you often make many corrections when pointing. The Angle Mouse\r\nwill make small targets easier to acquire by lowering the mouse speed as you try to place\r\nyour cursor inside them.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void lklblWayTooMany_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Choose this preset if you feel you make way too many corrections when pointing. The Angle\r\nMouse will dramatically lower the mouse speed as you try to place your mouse cursor inside\r\ntargets on the screen.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        #endregion

        #region Explanatory Hyperlink Labels: Custom Tab

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblActivate_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this, 
                new Point(Right, Top),
                false,
                "When checked, the Angle Mouse is active, and the mouse speed, known as gain,\r\nwill be adjusted as the mouse moves to improve pointing performance.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblUsePA_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Pointer acceleration is a Windows feature that is enabled with the\r\nEnhance pointer precision checkbox on the mouse control panel.\r\nIt makes short movements easy without making long movements\r\nmore difficult. This feature's interaction with the Angle Mouse has\r\nnot been formally studied.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblDistanceThreshold_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "This threshold indicates the distance, in pixels, that the mouse must move before\r\neach additional angle is calculated. Increasing this value will cause fewer angles\r\nto be calculated during movement, making gain changes less frequent. Decreasing\r\nthis value will cause more angles to be calculated during movement, making the\r\nAngle Mouse more sensitive.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblQueueLength_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "The queue length indicates the number of moves remembered by the Angle Mouse.\r\nSince two moves are required to form one angle, this value also indicates the number\r\nof angles remembered, which is always one less than the number of moves. Increasing\r\nthis value could introduce some lag. Decreasing this value could cause the Angle\r\nMouse to fail to change the gain advantageously.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMaxGauss_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "The maximum Gaussian standard deviation determines how flat the angle weighting\r\ncurve becomes. The higher this value, the flatter the bottom curve, giving equal\r\nweight to all angles in the queue. This weighting kicks in when movement is most\r\ndivergent.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMinGauss_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "The minimum Gaussian standard deviation determines how peaked the angle weighting\r\ncurve becomes. The lower this value, the higher the top curve, giving more weight to\r\nonly the most recent angles. This weighting kicks in when movement is most coherent.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMaxGain_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "The maximum mouse speed (1-20) indicates how fast the mouse cursor is allowed\r\nto move when movement is most coherent. This value is based on the pointer speed\r\nslider on the mouse control panel.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMinGain_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "The minimum mouse speed (1-20) indicates how slow the mouse cursor is allowed\r\nto move when movement is most divergent. This value is based on the pointer speed\r\nslider on the mouse control panel.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        #endregion

        #region Explanatory Hyperlink Labels: Output Tab

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblAngles_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the number of angles currently in the angle queue. The maximum\r\nvalue is governed by the movement queue length option.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMean_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the average, or most central, angle at which the cursor is currently\r\nmoving. Moving right is 0°, down is 90°, left is 180°, and up is 270°.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblDeviation_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the angular deviation, or spread of angles, at which the cursor is currently\r\nmoving. The more spread, the lower the mouse speed. The less spread, the higher\r\nthe mouse speed.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMaxDeviation_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the maximum possible angular deviation for the number of angles in\r\nthe queue.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblPercent_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the percentage of mouse speed retained based on the ratio of deviation\r\nto maximum deviation. The higher the percent, the greater the mouse speed. The\r\nlower the percent, the lower the mouse speed.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="sender">The sender of this event.</param>
        /// <param name="e">The arguments for this event.</param>
        private void lklblMouseSpeed_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            PositionableMessageBox.Show(this,
                new Point(Right, Top),
                false,
                "Indicates the current mouse speed setting, which is bound by the minimum and\r\nmaximum mouse speed options. When movement is coherent, mouse speed will\r\nbe higher. When movement is divergent, mouse speed will be lower.",
                ((LinkLabel) sender).Text,
                true,
                MessageBoxIcon.Information,
                true);
        }

        #endregion
    }
}
