﻿/**
 * Angle Mouse
 *
 *		Jacob O. Wobbrock, Ph.D.
 * 		The Information School
 *		University of Washington
 *		Mary Gates Hall, Box 352840
 *		Seattle, WA 98195-2840
 *		wobbrock@uw.edu
 *		
 * This software is distributed under the "New BSD License" agreement:
 * 
 * Copyright (c) 2009-2011, Jacob O. Wobbrock. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of the University of Washington nor the names of its 
 *      contributors may be used to endorse or promote products derived from 
 *      this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Jacob O. Wobbrock
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
**/
using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace AngleMouse
{
    [Serializable]
    public class Options : ISerializable
    {
        public bool IsActive; // whether or not the Angle Mouse is active
        public double DistanceThreshold; // movement distance threshold
        public int QueueLength; // number of moves N means N - 1 stored angles
        public int MaxGaussianStdev; // maximum Gaussian standard deviation (produces flattest curve for divergent movement)
        public int MinGaussianStdev; // minimum Gaussian standard deviation (produces peaked curve for coherent movement)
        public uint MaxGain; // up to 20
        public uint MinGain; // down to 1
        public bool UsePointerAcceleration; // whether or not to use Windows pointer acceleration

        /// <summary>
        /// Defaults.
        /// </summary>
        public Options(PresetType p)
        {
            this.IsActive = true;
            this.UsePointerAcceleration = true;
            this.DistanceThreshold = 8.0; // 8 px
            this.QueueLength = 17; // 16 angles remembered

            switch (p)
            {
                case PresetType.HardlyAny:
                    this.MinGaussianStdev = 3; // stdev when coherent
                    this.MaxGaussianStdev = 11; // stdev when divergent
                    this.MaxGain = 18u; // gain when coherent
                    this.MinGain = 4u; // gain when divergent
                    break;
                case PresetType.AFew:
                    this.MinGaussianStdev = 4; // stdev when coherent
                    this.MaxGaussianStdev = 15; // stdev when divergent
                    this.MaxGain = 16u; // gain when coherent
                    this.MinGain = 2u; // gain when divergent
                    break;
                case PresetType.Some:
                    this.MinGaussianStdev = 5; // stdev when coherent
                    this.MaxGaussianStdev = 17; // stdev when divergent
                    this.MaxGain = 14u; // gain when coherent
                    this.MinGain = 2u; // gain when divergent
                    break;
                case PresetType.Many:
                    this.MinGaussianStdev = 6; // stdev when coherent
                    this.MaxGaussianStdev = 19; // stdev when divergent
                    this.MaxGain = 12u; // gain when coherent
                    this.MinGain = 2u; // gain when divergent
                    break;
                case PresetType.WayTooMany:
                    this.MinGaussianStdev = 7; // stdev when coherent
                    this.MaxGaussianStdev = 21; // stdev when divergent
                    this.MaxGain = 10u; // gain when coherent
                    this.MinGain = 2u; // gain when divergent
                    break;
                default:
                    throw new ArgumentException("Invalid argument to Options constructor.");
            }
        }

        /// <summary>
        /// Copy constructor.
        /// </summary>
        /// <param name="toCopy">Options instance to copy.</param>
        public Options(Options toCopy)
        {
            this.IsActive = toCopy.IsActive;
            this.DistanceThreshold = toCopy.DistanceThreshold;
            this.QueueLength = toCopy.QueueLength;
            this.MaxGaussianStdev = toCopy.MaxGaussianStdev;
            this.MinGaussianStdev = toCopy.MinGaussianStdev;
            this.MaxGain = toCopy.MaxGain;
            this.MinGain = toCopy.MinGain;
            this.UsePointerAcceleration = toCopy.UsePointerAcceleration;
        }

        /// <summary>
        /// Deserialization constructor.
        /// </summary>
        /// <param name="info">Serialization information.</param>
        /// <param name="context">Streaming context. Currently unused.</param>
        public Options(SerializationInfo info, StreamingContext context)
        {
            this.IsActive = info.GetBoolean("IsActive");
            this.DistanceThreshold = info.GetDouble("DistanceThreshold");
            this.QueueLength = info.GetInt32("QueueLength");
            this.MaxGaussianStdev = info.GetInt32("MaxGaussianStdev");
            this.MinGaussianStdev = info.GetInt32("MinGaussianStdev");
            this.MaxGain = info.GetUInt32("MaxGain");
            this.MinGain = info.GetUInt32("MinGain");
            this.UsePointerAcceleration = info.GetBoolean("UsePointerAcceleration");
        }

        /// <summary>
        /// Called during serialization. Saves the data out for reading in later.
        /// </summary>
        /// <param name="info">Serialization information. Holds whatever properties we save into it.</param>
        /// <param name="context">Serialization context. Currently unused.</param>
        public void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            info.AddValue("IsActive", this.IsActive);
            info.AddValue("DistanceThreshold", this.DistanceThreshold);
            info.AddValue("QueueLength", this.QueueLength);
            info.AddValue("MaxGaussianStdev", this.MaxGaussianStdev);
            info.AddValue("MinGaussianStdev", this.MinGaussianStdev);
            info.AddValue("MaxGain", this.MaxGain);
            info.AddValue("MinGain", this.MinGain);
            info.AddValue("UsePointerAcceleration", this.UsePointerAcceleration);
        }

        /// <summary>
        /// Deep copy method.
        /// </summary>
        /// <returns>A deep copy of this Options instance.</returns>
        public Options Copy()
        {
            return new Options(this);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="o1"></param>
        /// <param name="o2"></param>
        /// <returns></returns>
        public static bool operator ==(Options o1, Options o2)
        {
            return
                o1.IsActive == o2.IsActive
                && o1.DistanceThreshold == o2.DistanceThreshold
                && o1.QueueLength == o2.QueueLength
                && o1.MaxGaussianStdev == o2.MaxGaussianStdev
                && o1.MinGaussianStdev == o2.MinGaussianStdev
                && o1.MaxGain == o2.MaxGain
                && o1.MinGain == o2.MinGain
                && o1.UsePointerAcceleration == o2.UsePointerAcceleration;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="o1"></param>
        /// <param name="o2"></param>
        /// <returns></returns>
        public static bool operator !=(Options o1, Options o2)
        {
            return
                o1.IsActive != o2.IsActive
                || o1.DistanceThreshold != o2.DistanceThreshold
                || o1.QueueLength != o2.QueueLength
                || o1.MaxGaussianStdev != o2.MaxGaussianStdev
                || o1.MinGaussianStdev != o2.MinGaussianStdev
                || o1.MaxGain != o2.MaxGain
                || o1.MinGain != o2.MinGain
                || o1.UsePointerAcceleration != o2.UsePointerAcceleration;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="obj"></param>
        /// <returns></returns>
        public override bool Equals(object obj)
        {
            if (obj is Options)
            {
                Options op = (Options) obj;
                return this == op;
            }
            return false;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public PresetType TestForPreset()
        {
            for (PresetType p = PresetType.HardlyAny; p <= PresetType.WayTooMany; p++)
            {
                Options op = new Options(p);
                if (this == op)
                    return p;
            }
            return PresetType.None;
        }

        #region Unused

        /// <summary>
        /// The maximum Gaussian standard deviation is 3x the standard deviation needed to just
        /// cover each slot (that is, just reach zero at the last angle to be weighted). It would 
        /// take an angle queue 3x as long for this Gaussian standard deviation to reach zero at
        /// the last slot.
        /// </summary>
        /// <remarks>
        /// A discretized Gaussian distribution reaches zero about 3 standard deviations from its center. So
        /// a full Gaussian curve would be an array of size 3 * stdev * 2 + 1. The size of the half-Gaussian 
        /// kernel will be therefore 3 * stdev + 1. So for n stored angles, we have m stored moves where
        /// m = n + 1, so stdev = (m - 2) / 3 for weighting to the last slot.
        /// </remarks>
        //private int GetMaxGaussianStdev(int queuelen)
        //{
        //    return queuelen - 2;
        //}

        /// <summary>
        /// The minimum Gaussian standard deviation is the standard deviation needed to just
        /// cover each slot (that is, just reach zero at the last angle to be weighted).
        /// </summary>
        /// <remarks>
        /// A discretized Gaussian distribution reaches zero about 3 standard deviations from its center. So
        /// a full Gaussian curve would be an array of size 3 * stdev * 2 + 1. The size of the half-Gaussian 
        /// kernel will be therefore 3 * stdev + 1. So for n stored angles, we have m stored moves where
        /// m = n + 1, so stdev = (m - 2) / 3 for weighting to the last slot.
        /// </remarks>  
        //private int GetMinGaussianStdev(int queuelen)
        //{
        //    return (queuelen - 2) / 3;
        //}

        #endregion
    }
}